/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Enrique Robledo Arnuncio, for digitalSTROM AG <era@tegulan.com>

    This file is part of ds485p.

    ds485p is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    ds485p is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ds485p. If not, see <http://www.gnu.org/licenses/>.
*/

#include "whitelist.h"

#include "base/log.h"

#include <ctype.h>
#include <sqlite3.h>
#include <string.h>
#include <unistd.h>

static sqlite3* db_handle = NULL;

/* One-time SQL statements without arguments. */

/* White-list table creation.
 *
 * Backwards-compatibility note:
 * If new columns need to be added to the table, they should be added to
 * this create statement, and a corresponding add_xxx_column_sql statement
 * should be executed in the attempt_schema_upgrade function below.
 */
static const char* create_sql =
    "create table if not exists autoconnect_whitelist("
    "dsuid varchar(35),"
    "noauto smallint,"
    "autoconnect smallint,"
    "name varchar(64),"
    "primary key (dsuid));";

static const char* add_name_column_sql =
    "alter table autoconnect_whitelist add column name varchar(64);";

/* SQL statements to implement whitelist_insert_and_get_autoconnect.
 * These may be executed many times, so we pre-compile each of them into
 * a sqlite3_stmt.
 *
 * See the comments in run_insert_and_get_autoconnect_statements below for
 * a description on how these statements get combined into transactions.
 */
static sqlite3_stmt* begin_stmt = NULL;
static const char* begin_sql = "begin exclusive transaction;";

static sqlite3_stmt* commit_stmt = NULL;
static const char* commit_sql = "commit transaction;";

#define STRING(val) #val
#define SQLITE_PARAM(idx) "?"STRING(idx)

#define SELECT_DSUID_ARG 1
#define SELECT_NOAUTO_COLUMN 0
#define SELECT_AUTOCONNECT_COLUMN 1
#define SELECT_NAME_COLUMN 2
static sqlite3_stmt* select_stmt = NULL;
static const char* select_sql =
    "select noauto, autoconnect, name from autoconnect_whitelist"
    " where dsuid = " SQLITE_PARAM(SELECT_DSUID_ARG) ";";

#define INSERT_DSUID_ARG 1
#define INSERT_NOAUTO_ARG 2
#define INSERT_AUTOCONNECT_ARG 3
#define INSERT_NAME_ARG 4
static sqlite3_stmt* insert_stmt = NULL;
static const char* insert_sql =
    "insert into autoconnect_whitelist (dsuid, noauto, autoconnect, name) values ("
    SQLITE_PARAM(INSERT_DSUID_ARG) ", "
    SQLITE_PARAM(INSERT_NOAUTO_ARG) ", "
    SQLITE_PARAM(INSERT_AUTOCONNECT_ARG) ", "
    SQLITE_PARAM(INSERT_NAME_ARG) ");";

#define UPDATE_DSUID_ARG 1
#define UPDATE_NOAUTO_ARG 2
#define UPDATE_NAME_ARG 3
static sqlite3_stmt* update_stmt = NULL;
static const char* update_sql =
    "update autoconnect_whitelist set"
    " noauto = " SQLITE_PARAM(UPDATE_NOAUTO_ARG) ","
    " name = " SQLITE_PARAM(UPDATE_NAME_ARG)
    " where dsuid = " SQLITE_PARAM(UPDATE_DSUID_ARG) ";";

/* Minimal select statement to retrieve the autoconnect flag only */
#define SELECT2_DSUID_ARG 1
#define SELECT2_AUTOCONNECT_COLUMN 0
static sqlite3_stmt* select2_stmt = NULL;
static const char* select2_sql =
    "select autoconnect from autoconnect_whitelist"
    " where dsuid = " SQLITE_PARAM(SELECT_DSUID_ARG) ";";

#undef SQLITE_PARAM
#undef STRING

static int dsuid_to_string_upper(const dsuid_t* dsuid,
                                 DSUID_STR(dsuid_str)) {
    const int res = dsuid_to_string(dsuid, dsuid_str);
    if (res != DSUID_RC_OK) {
        return res;
    }
    size_t ctr;
    for (ctr = 0; ctr < DSUID_STR_LEN - 1; ++ctr) {
        dsuid_str[ctr] = toupper(dsuid_str[ctr]);
    }
    return res;
}

static void close_database() {
    /* Note that passing a NULL to the finalize function is a no-op */
    sqlite3_finalize(update_stmt);
    sqlite3_finalize(insert_stmt);
    sqlite3_finalize(select_stmt);
    sqlite3_finalize(begin_stmt);
    sqlite3_finalize(commit_stmt);
    sqlite3_close_v2(db_handle);
    db_handle = NULL;
}

/* Backwards compatibility mechanism.
 *
 * New columns may be added to the white-list schema while preserving
 * compatibility with existing white-list files.
 *
 * Unfortunately, there is no clean way to query the schema of an existing
 * database in sqlite3. The simplest mechanism is the table_info PRAGMA
 * statement, which is not guaranteed to exist in future versions of sqlite. And
 * the sqlite_master table mechanism is very cumbersome.
 *
 * Instead of verifying if the schema is complete, we wait until a SQL statement
 * generates an error in prepare_statement, and then try to add all the
 * potentially missing columns with attempt_schema_upgrade.
 */
static bool attempt_schema_upgrade() {
    bool upgraded = false;
    if (sqlite3_exec(db_handle, add_name_column_sql, NULL, NULL, NULL) == SQLITE_OK) {
        report(LOG_INFO, "Added \"name\" column to the whitelist\n");
        upgraded = true;
    }
    return upgraded;
}

static bool prepare_statement(const char* sql, sqlite3_stmt** stmt) {
    int res = sqlite3_prepare_v2(db_handle, sql, strlen(sql), stmt, NULL);
    if (res == SQLITE_ERROR) {
        if (attempt_schema_upgrade()) {
            /* Some columns were added. We retry. */
            res = sqlite3_prepare_v2(db_handle, sql, strlen(sql), stmt, NULL);
        }
    }
    if (res != SQLITE_OK) {
        report(LOG_ERR, "SQL prepare error: %s\n", sqlite3_errstr(res));
        report(LOG_ERR, "In statement: %s\n", sql);
        close_database();
        return false;
    }
    return true;
}

bool whitelist_init(const char* sqlite_filename) {
    int res = sqlite3_open_v2(
        sqlite_filename,
        &db_handle,
        SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX,
        NULL);        /* Name of VFS module to use */
    if (res != SQLITE_OK) {
        db_handle = NULL;
        report(LOG_ERR, "Could not open or create the vdsm whitelist: %s\n",
               sqlite_filename);
        report(LOG_ERR, " sqlite3 error: %s\n", sqlite3_errstr(res));
        return false;
    }
    char* error = NULL;
    if (sqlite3_exec(db_handle, create_sql, NULL, NULL, &error) != SQLITE_OK) {
        report(LOG_ERR, "Could not create the whitelist table: %s\n",
                    error ? error : "Unknown error.");
        sqlite3_free(error);
        close_database();
        return false;
    }
    if (!prepare_statement(begin_sql, &begin_stmt)) {
        return false;
    }
    if (!prepare_statement(commit_sql, &commit_stmt)) {
        return false;
    }
    if (!prepare_statement(select_sql, &select_stmt)) {
        return false;
    }
    if (!prepare_statement(insert_sql, &insert_stmt)) {
        return false;
    }
    if (!prepare_statement(update_sql, &update_stmt)) {
        return false;
    }
    if (!prepare_statement(select2_sql, &select2_stmt)) {
        return false;
    }
    return true;
}

bool whitelist_exists() {
    return db_handle != NULL;
}

static bool run_begin() {
    sqlite3_reset(begin_stmt);
    int res;
    while (1) {
        res = sqlite3_step(begin_stmt);
        if (res == SQLITE_BUSY) {
            sleep(1);
            continue;
        }
        break;
    }
    return res == SQLITE_DONE;
}

void whitelist_free_fields(vdsm_whitelist_fields_t* fields) {
    free((void*)fields->name);
}

/* The pointers returned in "fields" need to be freed with whitelist_free_fields.
 * The returned pointers may be null if the corresponding columns in the whitelist
 * database are not set (SQL NULL).
 */
static bool run_select(const char* dsuid,
                       bool *found,
                       vdsm_whitelist_fields_t* fields) {
    fields->name = NULL;
    sqlite3_reset(select_stmt);
    sqlite3_clear_bindings(select_stmt);
    int res = sqlite3_bind_text(
        select_stmt, SELECT_DSUID_ARG, dsuid, strlen(dsuid), SQLITE_STATIC);
    if (res != SQLITE_OK) {
        return false;
    }
    res = sqlite3_step(select_stmt);
    if (res == SQLITE_DONE) {
        *found = false;
    } else if (res == SQLITE_ROW) {
        *found = true;
        fields->noauto =
            sqlite3_column_int(select_stmt, SELECT_NOAUTO_COLUMN);
        fields->autoconnect =
            sqlite3_column_int(select_stmt, SELECT_AUTOCONNECT_COLUMN);
        char* name = (char*)sqlite3_column_text(select_stmt, SELECT_NAME_COLUMN);
        if (name != NULL)
            fields->name = strdup(name);
        res = sqlite3_step(select_stmt);
    }
    return res == SQLITE_DONE;
}

static bool run_update(const char* dsuid, bool noauto, const char* name) {
    sqlite3_reset(update_stmt);
    sqlite3_clear_bindings(update_stmt);
    int res = sqlite3_bind_text(
        update_stmt, UPDATE_DSUID_ARG, dsuid, strlen(dsuid), SQLITE_STATIC);
    if (res == SQLITE_OK) {
        res = sqlite3_bind_int(update_stmt, UPDATE_NOAUTO_ARG, noauto);
    }
    if (res == SQLITE_OK) {
        res = sqlite3_bind_text(
            update_stmt,
            UPDATE_NAME_ARG,
            name,
            name ? strlen(name) : 0,
            SQLITE_STATIC);
    }
    if (res != SQLITE_OK) {
        return false;
    }
    res = sqlite3_step(update_stmt);
    return res == SQLITE_DONE;
}

static bool run_insert(const char* dsuid,
                       const vdsm_whitelist_fields_t* fields) {
    sqlite3_reset(insert_stmt);
    sqlite3_clear_bindings(insert_stmt);
    int res = sqlite3_bind_text(
        insert_stmt, INSERT_DSUID_ARG, dsuid, strlen(dsuid), SQLITE_STATIC);
    if (res == SQLITE_OK) {
        res = sqlite3_bind_int(insert_stmt, INSERT_NOAUTO_ARG, fields->noauto);
    }
    if (res == SQLITE_OK) {
        res = sqlite3_bind_int(insert_stmt, INSERT_AUTOCONNECT_ARG, fields->autoconnect);
    }
    if (res == SQLITE_OK) {
        res = sqlite3_bind_text(
            insert_stmt,
            INSERT_NAME_ARG,
            fields->name,
            fields->name ? strlen(fields->name) : 0,
            SQLITE_STATIC);
    }
    if (res != SQLITE_OK) {
        return false;
    }
    res = sqlite3_step(insert_stmt);
    return res == SQLITE_DONE;
}

static bool run_commit() {
    sqlite3_reset(commit_stmt);
    int res = sqlite3_step(commit_stmt);
    return res == SQLITE_DONE;
}

/* The SQL transaction that implements whitelist_insert_and_get_autoconnect starts with a
 * "select" statement to test the presence of the dsuid.  After that, the sequence of
 * statements is different depending on whether an entry for the dsuid already exists,
 * and whether noauto has changed.
 *
 * SQL Statements when the dsuid is not present in the whitelist:
 *
 *   begin exclusive transaction;
 *   select noauto, autoconnect, name from autoconnect_whitelist
 *       where dsuid = ?1;
 *   insert into autoconnect_whitelist
 *       (dsuid, noauto, autoconnect, name) values(?1, ?2, ?3, ?4);
 *   commit transaction
 *
 * SQL Statements when the dsuid is already present, and noauto has not changed:
 *
 *   begin exclusive transaction;
 *   select noauto, autoconnect, name from autoconnect_whitelist
 *       where dsuid = ?1;
 *   commit transaction
 *
 * SQL Statements when the dsuid is already present, and noauto or name has
 * changed:
 *
 *   begin exclusive transaction;
 *   select noauto, autoconnect, name from autoconnect_whitelist
 *       where dsuid = ?1;
 *   update autoconnect_whitelist set noauto = ?2, name = ?3
 *       where dsuid = $1;
 *   commit transaction
 */
static bool run_insert_and_get_autoconnect_statements(
    const char* dsuid, const vdsm_whitelist_fields_t* fields, bool* autoconnect) {
    if (!run_begin()) {
        return false;
    }
    bool found = false;
    vdsm_whitelist_fields_t db_fields;
    if (!run_select(dsuid, &found, &db_fields)) {
        whitelist_free_fields(&db_fields);
        return false;
    }
    if (found) {
        bool name_changed;
        if (db_fields.name != NULL && fields->name != NULL) {
            name_changed = (strcmp(db_fields.name, fields->name) != 0);
        } else {
            name_changed = (db_fields.name != NULL || fields->name != NULL);
        }
        if (name_changed || db_fields.noauto != fields->noauto) {
            if (!run_update(dsuid, fields->noauto, fields->name)) {
                whitelist_free_fields(&db_fields);
                return false;
            }
        }
        *autoconnect = db_fields.autoconnect;
    } else {
        if (!run_insert(dsuid, fields)) {
            whitelist_free_fields(&db_fields);
            return false;
        }
        *autoconnect = fields->autoconnect;
    }
    whitelist_free_fields(&db_fields);
    if (!run_commit()) {
        return false;
    }
    return true;
}

bool whitelist_insert_and_get_autoconnect(const dsuid_t* dsuid,
                                          const vdsm_whitelist_fields_t* fields) {
    if (!db_handle) {
        return fields->autoconnect;
    }
    DSUID_STR(dsuid_str);
    if (dsuid_to_string_upper(dsuid, dsuid_str) != DSUID_RC_OK) {
        report(LOG_ERR, "Invalid DSUID.\n");
        return fields->autoconnect;
    }
    bool autoconnect = fields->autoconnect;
    if (!run_insert_and_get_autoconnect_statements(dsuid_str, fields, &autoconnect)) {
        run_commit();
        close_database();
    }
    return autoconnect;
}

bool whitelist_get_autoconnect(const dsuid_t* dsuid) {
    sqlite3_reset(select2_stmt);
    sqlite3_clear_bindings(select2_stmt);
    DSUID_STR(dsuid_str);
    if (dsuid_to_string_upper(dsuid, dsuid_str) != DSUID_RC_OK) {
        report(LOG_ERR, "Invalid DSUID.\n");
        return false;
    }
    dsuid_to_string_upper(dsuid, dsuid_str);
    int res = sqlite3_bind_text(
        select2_stmt, SELECT2_DSUID_ARG, dsuid_str, strlen(dsuid_str), SQLITE_STATIC);
    if (res != SQLITE_OK) {
        return false;
    }
    res = sqlite3_step(select2_stmt);
    if (res != SQLITE_ROW) {
        return false;
    }
    const bool autoconnect =
        sqlite3_column_int(select2_stmt, SELECT2_AUTOCONNECT_COLUMN);
    res = sqlite3_step(select2_stmt);
    if (res != SQLITE_DONE) {
        return false;
    }
    return autoconnect;
}

bool whitelist_get(const dsuid_t* dsuid, vdsm_whitelist_fields_t* fields) {
    if (!db_handle) {
        return fields->autoconnect;
    }
    DSUID_STR(dsuid_str);
    if (dsuid_to_string_upper(dsuid, dsuid_str) != DSUID_RC_OK) {
        report(LOG_ERR, "Invalid DSUID.\n");
        return fields->autoconnect;
    }
    bool found = false;
    if (!run_select(dsuid_str, &found, fields)) {
        run_commit();
        close_database();
        return false;
    }
    return found;
}

bool whitelist_import_legacy(vdsm_legacy_whitelist_hash_t **legacy_whitelist) {
    if (!db_handle) {
        return false;
    }
    vdsm_legacy_whitelist_hash_t *entry, *tmp;
    vdsm_whitelist_fields_t fields;
    fields.noauto = false;
    fields.autoconnect = true;
    HASH_ITER(hh, *legacy_whitelist, entry, tmp) {
        DSUID_STR(dsuid_str);
        dsuid_to_string(&entry->dsuid, dsuid_str);
        run_insert(dsuid_str, &fields);
    }
    return true;
}
